use std::{
    error::Error as StdErr,
    fmt::Display,
    fs::{self, OpenOptions},
    io::{self, Write},
    path::{Path, PathBuf},
    process::Command,
};

use clap::{command, Parser};
use serde::Deserialize;
use shellexpand::LookupError;

const CONFIG_DIR: &'static str = "launchd";
const CONFIG_NAME: &'static str = "config.toml";

#[derive(Debug)]
enum ErrorKind {
    ConfigNotFound,
    CantReadConfig,
    DirNotFound,
    CantReadDir,
    InvalidInput,
    NoInputGiven,
    CantExecute,
}

#[derive(Debug)]
struct Error {
    kind: ErrorKind,
    source: Option<Box<dyn StdErr>>,
}

impl Error {
    pub fn new(kind: ErrorKind, source: Option<Box<dyn StdErr>>) -> Self {
        Error { kind, source }
    }
}

impl From<ErrorKind> for Error {
    fn from(kind: ErrorKind) -> Self {
        Self::new(kind, None)
    }
}

impl<T: StdErr + 'static> From<LookupError<T>> for Error {
    fn from(value: LookupError<T>) -> Self {
        Self::new(ErrorKind::CantReadDir, Some(Box::new(value)))
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.kind {
            ErrorKind::ConfigNotFound => write!(f, "could not find config file"),
            ErrorKind::CantReadConfig => write!(f, "could not read config file"),
            ErrorKind::DirNotFound => write!(f, "could not find directory"),
            ErrorKind::CantReadDir => write!(f, "could not read directory"),
            ErrorKind::InvalidInput => write!(f, "given input is not valid"),
            ErrorKind::NoInputGiven => write!(f, "input canceled by user"),
            ErrorKind::CantExecute => write!(f, "couldn't launch executable"),
        }
    }
}

impl StdErr for Error {}

#[derive(Deserialize, Debug)]
pub struct Config {
    pub executable_path: String,
    pub wad_path: String,
    pub default_wad: Option<String>,
}

fn touch<P: AsRef<Path>>(path: P) -> io::Result<()> {
    match OpenOptions::new().create(true).write(true).open(path) {
        Ok(_) => Ok(()),
        Err(e) => Err(e),
    }
}

fn make_config_path() -> Option<PathBuf> {
    home::home_dir()
        .map(|p| p.join(".config"))
        .filter(|p| p.exists())
        .map(|p| p.join(CONFIG_DIR))
        .filter(|p| fs::create_dir_all(p).is_ok())
        .map(|p| p.join(CONFIG_NAME))
        .filter(|p| touch(p).is_ok())
}

fn config_path<P: AsRef<Path>>(path: &Option<P>) -> Result<PathBuf, Error> {
    if let Some(path) = path {
        let path = path.as_ref();
        if path.is_file() {
            Ok(path.to_path_buf())
        } else if path.is_dir() {
            config_path(&Some(path.join(CONFIG_NAME)))
        } else {
            Err(ErrorKind::ConfigNotFound.into())
        }
    } else {
        config_path(&make_config_path())
    }
}

fn read_config<P: AsRef<Path>>(path: P) -> Result<String, Error> {
    fs::read_to_string(path.as_ref())
        .map_err(|e| Error::new(ErrorKind::CantReadConfig, Some(Box::new(e))))
}

fn parse_config<'a, S: Into<&'a str>>(s: S) -> Result<Config, Error> {
    toml::from_str(s.into()).map_err(|e| Error::new(ErrorKind::CantReadConfig, Some(Box::new(e))))
}

fn scan_wads<P: AsRef<Path>>(path: P) -> Result<Vec<PathBuf>, Error> {
    let path = path.as_ref();

    if path.is_dir() {
        match fs::read_dir(path) {
            Ok(v) => Ok(v.filter_map(|x| x.ok()).map(|x| x.path()).collect()),
            Err(e) => Err(Error::new(ErrorKind::CantReadDir, Some(Box::new(e)))),
        }
    } else {
        Err(Error::new(ErrorKind::DirNotFound, None))
    }
}

fn prompt_wads<P: AsRef<Path>>(wads: &[P]) -> Result<&str, Error> {
    let mut wads = wads
        .iter()
        .filter_map(|p| p.as_ref().file_name())
        .filter_map(|p| p.to_str())
        .enumerate();

    for (i, wad) in wads.clone() {
        println!("{i}: {wad}");
    }

    print!("Please pick an option: ");
    std::io::stdout().flush().unwrap();
    let mut input = String::new();

    let selection = match io::stdin().read_line(&mut input) {
        Ok(_) if input == "q" => Err(ErrorKind::NoInputGiven.into()),
        Ok(0) => Err(ErrorKind::NoInputGiven.into()),
        Ok(_) => Ok(input),
        Err(e) => Err(Error::new(ErrorKind::InvalidInput, Some(Box::new(e)))),
    }?;

    let selection = selection
        .trim()
        .parse::<usize>()
        .map_err(|e| Error::new(ErrorKind::InvalidInput, Some(Box::new(e))))?;

    match wads.find(|(i, _)| *i == selection).map(|(_, wad)| wad) {
        Some(v) => Ok(v),
        None => Err(Error::new(ErrorKind::InvalidInput, None)),
    }
}

fn select_wad<'a, P: AsRef<Path>>(wad: &str, wads: &'a [P]) -> Option<&'a Path> {
    wads.iter()
        .map(|p| p.as_ref().to_str())
        .filter_map(|x| x)
        .find(|p| p.contains(wad))
        .map(Path::new)
}

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Arguments {
    #[arg(short, long)]
    config_path: Option<String>,
    #[arg(short = 'd', long, default_value_t = false)]
    set_default: bool,
    wad: Option<String>,
}

fn main() -> Result<(), Error> {
    let args = Arguments::parse();
    let config = config_path(&args.config_path)
        .and_then(read_config)
        .and_then(|s| parse_config(s.as_str()))?;

    let wads = shellexpand::full(&config.wad_path)
        .map_err(Into::into)
        .and_then(|p| scan_wads(p.to_string()))?;

    let wad = if args.wad.is_some() {
        args.wad
    } else {
        println!("using directory {}:", config.wad_path);
        prompt_wads(&wads).map(str::to_string).ok()
    };

    if let Some(wad) = wad {
        if let Some(selected) = select_wad(&wad, wads.as_slice()) {
            let exec = shellexpand::full(&config.executable_path)?.to_string();

            Command::new(&exec)
                .arg(selected.as_os_str())
                .spawn()
                .map_err(|e| Error::new(ErrorKind::CantExecute, Some(Box::new(e))))?;
        }
    }

    Ok(())
}
